//---------------------------------------------------------------------------
#ifndef AffineTransformation_templCodeH
#define AffineTransformation_templCodeH
//---------------------------------------------------------------------------
//#include "GlobalDefinitions.h"
//---------------------------------------------------------------------------
#include "AffineTransformation.h"
//---------------------------------------------------------------------------
#include <math.h>
#include <string>
//---------------------------------------------------------------------------

using namespace std;


// ---------------------------------------------------------------------------
//
//                                 StaticMatrix
//
// ---------------------------------------------------------------------------

/************************************************************************/
/* Affine3d                                                             */
/************************************************************************/

inline Affine3d::Affine3d(double *velocity)
{

}

inline Vector3d& Affine3d::operator[](const unsigned &index)
{
	return v[index];
}
inline const Vector3d& Affine3d::operator[](const unsigned &index) const
{
	return v[index];
}

inline Affine3d Affine3d::operator+(const Affine3d &op) const 
{
	Affine3d result;
	result.v[0] = v[0] + op.v[0];
	result.v[1] = v[1] + op.v[1];
	result.v[2] = v[2] + op.v[2];
	result.v[3] = v[3] + op.v[3];
	return result;
}


inline Affine3d Affine3d::operator-(const Affine3d &op) const 
{
	Affine3d result;
	result.v[0] = v[0] - op.v[0];
	result.v[1] = v[1] - op.v[1];
	result.v[2] = v[2] - op.v[2];
	result.v[3] = v[3] - op.v[3];
	return result;
}


inline Affine3d Affine3d::operator-() const 
{
	Affine3d result;
	result.v[0] = -v[0];
	result.v[1] = -v[1];
	result.v[2] = -v[2];
	result.v[3] = -v[3];
	return result;
}


inline Affine3d Affine3d::operator*(const double &s) const 
{
	Affine3d result;
	result.v[0] = v[0] * s;
	result.v[1] = v[1] * s;
	result.v[2] = v[2] * s;
	result.v[3] = v[3] * s;

	return result;
}


inline Affine3d Affine3d::operator*(const Affine3d &op) const 
{
	Affine3d result;
	result.v[0] = v[0] + v[1]*op.v[0][0] + v[2]*op.v[0][1] + v[3]*op.v[0][2];
	for (int k = 1; k < 4; k++)
	{
		result.v[k] = v[1]*op.v[k][0] + v[2]*op.v[k][1] + v[3]*op.v[k][2];
	}

	return result;
}

inline Affine3d Affine3d::operator/(const double &s) const 
{
	Affine3d result;
	result.v[0] = v[0] / s;
	result.v[1] = v[1] / s;
	result.v[2] = v[2] / s;
	result.v[3] = v[3] / s;
	return result;
}


inline Affine3d Affine3d::operator+=(const Affine3d &op) 
{
	v[0] += op.v[0];
	v[1] += op.v[1];
	v[2] += op.v[2];
	v[3] += op.v[3];
	return *this;
}


inline Affine3d Affine3d::operator-=(const Affine3d &op) 
{
	v[0] -= op.v[0];
	v[1] -= op.v[1];
	v[2] -= op.v[2];
	v[3] -= op.v[3];
	return *this;
}


inline Affine3d Affine3d::operator*=(const double &op) 
{
	v[0] *= op;
	v[1] *= op;
	v[2] *= op;
	v[3] *= op;
	return *this;
}

inline Affine3d Affine3d::operator/=(const double &op)
{
	v[0] /= op;
	v[1] /= op;
	v[2] /= op;
	v[3] /= op;
	return *this;
}


inline Affine3d Affine3d::operator=(const Affine3d &op) 
{
	v[0] = op.v[0];
	v[1] = op.v[1];
	v[2] = op.v[2];
	v[3] = op.v[3];
	return *this;
}



inline Vector3d Affine3d::operator*(const Vector3d &vec) const 
{
	return v[0] + v[1]*vec[0] + v[2]*vec[1] + v[3]*vec[2];
}


inline Affine3d::Affine3d() 
{
	initialize();
}


inline void Affine3d::initialize()
{
	setZero();
	v[1][0] = 1.0;
	v[2][1] = 1.0;
	v[3][2] = 1.0;
}


inline void Affine3d::setZero()
{
	v[0][0] = v[0][1] = v[0][2] = 0;
	v[1][0] = v[1][1] = v[1][2] = 0;
	v[2][0] = v[2][1] = v[2][2] = 0;
	v[3][0] = v[3][1] = v[3][2] = 0;
}

inline double Affine3d::det()
{
	return v[1][0]*(v[2][1]*v[3][2] - v[3][1]*v[2][2])
		+ v[2][0]*(v[3][1]*v[1][2]-v[1][1]*v[3][2])
		+ v[3][0]*(v[1][1]*v[2][2]-v[2][1]*v[1][2]);
}

inline Affine3d	Affine3d::inverse()
{
	Affine3d	result;

	for (int i = 0; i < 3; i++){
		for (int j = 0; j < 3; j++){
			result.v[i+1][j] = v[j+1][i];
		}
	}
	result.v[0] = -(result.v[1]*v[0][0]
	+ result.v[2]*v[0][1]
	+ result.v[3]*v[0][2]);
	return result;
}

#endif
